/* dbgmem.c */
#include <stdlib.h>
#include <iostream.h>
#include <string.h>
#include <ctype.h>

#include "dbgmem.h"

// set the memory with header and debug info
void set_mem(char *ptr, unsigned size,short flag)
{
	headerp hptr;
	unsigned *eptr;

	/* given the pointer to the start of mem, assign the magic numbers
	and the size. */
	hptr = (headerp) ptr;
	hptr->magic = flag ? MAGIC_MALLOC : MAGIC_FREE;	
	hptr->size = size;
	eptr = (unsigned *) ((char *)ptr + END_OFFSET(size));
	*eptr = flag ? MAGIC_MALLOC : MAGIC_FREE;
}

// fill user mem with a color
void fill_mem(char *ptr, unsigned numbytes,short color)
{
	/* given the pointer to the usable area of memory,
	   color the memory as either unused, or free */
	char c1, c2;
	int i=0;
	c1 = color ? MALLOC_COLOR1 : FREE_COLOR1;
	c2 = color ? MALLOC_COLOR2 : FREE_COLOR2;
		
	for (i=1; i <= numbytes; i++)
	{
		*ptr++ = (i % 2) ? c1 : c2;
	}
}

// get "raw" system memory
headerp get_mem(unsigned size)
{
	unsigned tsize=0,bsize=0;
	
	char *ptr=0;
	
#ifdef DEBUGIT
	cout << "In get_mem, requesting " << size << " bytes" << endl;
#endif		
	tsize = TOT_BYTES(size);
	
#ifdef DEBUGIT
	cout << "After size adjustment, requesting " << tsize << endl;;
#endif		

	for (bsize = TOT_BYTES(DEFAULT_BLOCK_SIZE); ; bsize >>= 1)
	{
		if (bsize < tsize)
			bsize = tsize;
			
		if ( (ptr = (char *)malloc(bsize)) != 0)
			break;
		else if (bsize == tsize)
			return(0);
	}
	
#ifdef DEBUGIT
	cout << "System has memory! Putting on free list!" << endl;
#endif		
	// update the stats
	stats.current_heap_size += bsize;
	stats.used_blocks++; /* have to increment because free 
	                        will decrement */
	stats.used_bytes += USABLE_BYTES(bsize);
#ifdef DEBUGIT
	cout << "Current used_bytes is " << stats.used_bytes << endl;
#endif		

	// set the memory, i.e. give the block structure
	set_mem(ptr,USABLE_BYTES(bsize),MALLOCFLAG); 
	dbg_free(ret_addr(ptr)); /* free the new block to attach 
	                            it to the free list */	
	return(lfree); // return the block just freed
}

int check_magic(headerp ptr, short flag)
{
	int head=0,foot=0;
				
	head = flag? (ptr->magic == MAGIC_MALLOC) : 
	             (ptr->magic == MAGIC_FREE);
		
	foot = flag? (end_magic(ret_addr(ptr)) == MAGIC_MALLOC) :
						(end_magic(ret_addr(ptr)) == MAGIC_FREE);
	
	return ((head&&foot));
}		

/* check free list traverses the free-list, checks for corruption
and returns the number of blocks in the free-list and the
largest unallocated block. */
int check_free_list(unsigned *largest)
{
	headerp traverse;
	unsigned biggest=0;
	int cnt=0;
	
	if (!lfree)
	{
		*largest = 0;
		return(0);
	}
	
	for (traverse = freelist.next; !(traverse == &freelist); 
	                                 traverse = traverse->next)
	{
		if (!check_magic(traverse,FREEFLAG))
		{
			cout << 
			 "check_free_list: BAD MAGIC NUMBER! CORRUPTED HEAP!\n";
			return(0);
		} 		

		cnt++;
		if (traverse->size > biggest)
			biggest = traverse->size;
	}
	
	*largest = biggest;
	return(cnt);
}
			
/* check free list traverses the free-list, checks for corruption
and returns the number of blocks in the free-list and the
largest unallocated block. */
void print_free_list(void)
{
	headerp traverse;
	int cnt=0;
	
	if (!lfree)
		return;

	cout << "FREELIST\n";
	for (traverse = freelist.next; !(traverse == &freelist); 
	     traverse = traverse->next)
	{
		if (!check_magic(traverse,FREEFLAG))
		{
			cout << 
			"check_free_list: BAD MAGIC NUMBER! CORRUPTED HEAP!\n";
			return;
		} 		

		cnt++;		
		cout << "block[" << cnt << "]->size(" << traverse->size << ") ";		
		if (!(cnt % 3)) cout << endl;
	}
	cout << endl;
}			

void memory_map(void)
{
	headerp traverse,inbetween;
	int cnt=0;
	int blocks_seen=0;
	
	if (!lfree)
		return;

	cout << "MEMORY MAP" << endl;
	for (traverse = freelist.next; !(traverse == &freelist); 
	     traverse = traverse->next)
	{
		cnt++;
		cout << "[" << cnt << "]" << ret_addr(traverse) 
		     << ": FREE (" << traverse->size << ") ";
		if (!(cnt % 3))
			cout << endl;
			
		inbetween = (headerp) ( (char *) traverse 
		                      + TOT_BYTES(traverse->size));
		if (inbetween != traverse->next)
		{	
			while ( (inbetween < traverse->next) ||
				(blocks_seen < stats.used_blocks) )
			{
				/* if we hit a non-malloced block, which could occur if
				there is only one free block, and when the heap is not
				contiguous. get out and go to the next free block. */
				if (inbetween->magic != MAGIC_MALLOC)
					break;
				else
				{
					if (end_magic(ret_addr(inbetween)) != MAGIC_MALLOC)
					{
						cout << "memory_map: CORRUPTED HEAP!\n";
						cout << "memory_map: BAD end magic! on block : ";
						cout << cnt << endl;
						return;
					}
					blocks_seen++;
				}
				
				cnt++;
				cout << "[" << cnt << "]" << ret_addr(inbetween)
				     << ":MALLOC(" << inbetween->size << ") ";
				if (!(cnt % 3))
					cout << endl;
	 			
	 			inbetween = (headerp) ( (char *) inbetween 
	 			            + TOT_BYTES(inbetween->size));
	 		} // end of while
	 	}
	 	
	 } // end of for
	 cout << endl;
}			

int check_heap(void)
{
	headerp traverse,inbetween;
	int blocks_seen=0;
	
	for (traverse = freelist.next; !(traverse == &freelist); 
	     traverse = traverse->next)
	{

		if (!check_magic(traverse,FREEFLAG))
			return(0);
			
		inbetween = (headerp) ( (char *) traverse 
		            + TOT_BYTES(traverse->size));
		if (inbetween != traverse->next)
		{	
			while ( (inbetween < traverse->next) ||
				(blocks_seen < stats.used_blocks) )
			{
				if (inbetween->magic != MAGIC_MALLOC)
					break;
				else
				{
					if (end_magic(ret_addr(inbetween)) != MAGIC_MALLOC)
						return(0);
					blocks_seen++;
				}
	 			
	 			inbetween = (headerp) ( (char *) inbetween 
	 			           + TOT_BYTES(inbetween->size));
	 		} // end of while
	 	}
	 } // end of for
	 
	return(1);
}			
	
void *dbg_malloc(unsigned num_bytes)
{
	headerp traverse, previous_ptr;
	
#ifdef DEBUGIT
	cout << "In dbg_malloc, requesting " << num_bytes << " bytes.\n";
#endif		
	if (num_bytes == 0)
	{
		cout << "dbg_malloc: FATAL - num_bytes is 0!\n";
		exit(0);
	}
	
	// check if num_bytes is outlandish
	if (num_bytes > DEBUG_BOUND)
	{
		cout << "dbg_malloc: FATAL - " << num_bytes
		     << " greater than upper bound of " << DEBUG_BOUND << endl;
		exit(0);
	}
	
	if ( (previous_ptr = lfree) == 0)
	{
		freelist.next = previous_ptr = lfree = &freelist;
		freelist.size = 0;
	}	
		
#ifdef DEBUGIT
	cout << "Looking for a block equal to "
	     << MULTIPLE_OF(num_bytes) << " bytes.\n";
	cout << "OR looking for a block greater than or equal to " 
	     << TOT_BYTES(num_bytes) << " bytes.\n";
#endif		
	// search for a memory block big enough.
	for (traverse = previous_ptr->next; ; previous_ptr = traverse, 
		 traverse = traverse->next)
	{
		if (traverse != &freelist)
		{
			if (!check_magic(traverse,FREEFLAG))
			{
				cout << 
				  "dbg_malloc: BAD MAGIC NUMBER! CORRUPTED HEAP!\n";
				return 0;
			} 
		}
		
		/* traverse->size is a count of the USABLE BYTES.
		 looking for first memory equal to or bigger.   */
		if ( (traverse->size == MULTIPLE_OF(num_bytes)) ||
			 (traverse->size >= TOT_BYTES(num_bytes)) ) 
		{
#ifdef DEBUGIT
			cout << "Found a block (" << traverse->size
			     << " bytes) on free list.\n";
#endif		
			if (traverse->size == MULTIPLE_OF(num_bytes)) /* exact match */
			{
				previous_ptr->next = traverse->next; /* delink from free list */
				set_mem((char *)traverse,traverse->size,MALLOCFLAG);
				if (DEBUG_LEVEL == DEBUG_FILL)
					fill_mem((char *)ret_addr(traverse), traverse->size, MALLOCFLAG);		
			}
			else
			{
				// reduce size of current block so we can strip off the end.
				traverse->size -= TOT_BYTES(num_bytes);
				// set the memory of the reduced free block
				set_mem((char *)traverse,traverse->size,FREEFLAG);
				traverse = (headerp) ( (char *) traverse + TOT_BYTES(traverse->size));
				// set the memory of the tail portion.
				set_mem((char *)traverse, MULTIPLE_OF(num_bytes),MALLOCFLAG);
				if (DEBUG_LEVEL == DEBUG_FILL)
					fill_mem((char *)ret_addr(traverse), traverse->size, MALLOCFLAG);		
			}
			
			lfree = previous_ptr; /* set previous ptr to the place before the
									just-malloced memory so that if that memory
									gets freed right away, we will be ready at
									that spot. */
			
			// update the stats
			stats.used_blocks++;
			stats.used_bytes += traverse->size;
			stats.tot_allocs++;
#ifdef DEBUGIT
	cout << "Current used_bytes is " << stats.used_bytes << endl;
#endif
			if (HeapBounds.firstMalloc)
			{
				HeapBounds.firstMalloc = 0;  // one shot deal
				HeapBounds.min = (void *)traverse;
				HeapBounds.max = (void *)traverse;
			}
			else
			{
				if ((void *)traverse < HeapBounds.min)
					HeapBounds.min = (void *)traverse;
				else if ((void *)traverse > HeapBounds.max)
					HeapBounds.max = (void *)traverse;
			}
			
			return(ret_addr(traverse));
			
		} // end of if
				
		if (traverse == lfree)
		{
#ifdef DEBUGIT
			cout << "Traversed entire free list. Asking System!\n";
#endif		
			if ( (traverse = get_mem(num_bytes)) == 0)
				return(0);
		}
		
	} // end of for
}

void dbg_free(void *ptr)
{
	headerp block_header=0,traverse;

	if (!ptr)
	{
		cout << "dbg_free: Attempt to free a 0 pointer!\n";
		return;
	}
	else
		block_header = (headerp) ((char *)ptr - START_BYTES);
		
	/* Before we dereference a pointer, we must attempt to make sure
	   that it is a valid pointer.  To do this we will use two portable
	   methods. */
	long bytepos = (long) block_header;  // convert address to a number
	/* address must be a multiple of the DEFAULT ALIGNMENT,
	   otherwise we know it does not point to a previously malloced
	   block. (since we always return an aligned pointer) */
	if (bytepos % DEFAULT_ALIGNMENT)
	{
			cout << "dbg_free: Trying to free an INVALID pointer!\n";
			return;	
	}
	
	/* At this point we know that the pointer could point to a
	   valid block within memory.  The next question is - does this
	   pointer fall within our heap? Our answer to this is to see
	   if it falls within the MIN and MAX pointer malloced so far.
	   This solution assumes that the heap is contiguous. */
	if ( ( (void *)block_header < HeapBounds.min ) &&
	     ( (void *)block_header > HeapBounds.max )  )
	{
			cout << "dbg_free: Trying to free a NON-HEAP pointer!\n";
			return;	
	}

#ifdef DEBUGIT
	cout << "In dbg_free. Size of block to free is "
		 << block_header->size << endl;
#endif		

	/****** *check to make sure the pointer passed in has
	a MALLOC_MAGIC ***** */
		
	if (block_header->magic != MAGIC_MALLOC)
	{
		if (block_header->magic == MAGIC_FREE)
		{
			cout << "dbg_free: Trying to free a freed pointer!\n";
			return;
		}
		else
		{	
			cout << "dbg_free: Trying to free a NON-MALLOCED pointer!\n";
			return;
		}
	}
	
	// update the stats
	stats.used_blocks--;
	stats.used_bytes -= block_header->size;
	stats.tot_frees++;
#ifdef DEBUGIT
	cout << "Current used_bytes is " << stats.used_bytes << endl;
#endif		
	
	/* find where this block fits in the list. Should be inbetween
	two elements in the list. */
	for (traverse = lfree; 	
	!(block_header > traverse && block_header < traverse->next); 
	traverse = traverse->next)
	{
		/* if traverse is greater then its next pointer, we are at the end
		of the list. AND block header must be either greater then the end
		or before the beginning. */
		if (traverse >= traverse->next && (block_header > traverse ||
						block_header < traverse->next))
							break;
	}
			
	/* if the end_of_this_block equals the start of the next,
	   connect to upper */
	if (((char *) block_header + TOT_BYTES(block_header->size)) == 
									(char *) traverse->next)
	{
		block_header->size += TOT_BYTES(traverse->next->size);
		block_header->next = traverse->next->next;
		set_mem((char *)block_header,block_header->size,FREEFLAG);
		if (DEBUG_LEVEL == DEBUG_FILL)
			fill_mem((char *)ret_addr(block_header), block_header->size, 
			          FREEFLAG);		
	}
	else
	{
		set_mem((char *)block_header,block_header->size,FREEFLAG);
		block_header->next = traverse->next;
		if (DEBUG_LEVEL == DEBUG_FILL)
			fill_mem((char *)ret_addr(block_header), block_header->size, 
			          FREEFLAG);		
	}
	
	/* If the end of the previous block equals this blocks address,
	   connect to lower */
	if (((char *) traverse + TOT_BYTES(traverse->size)) == 
										(char *) block_header)
	{
		traverse->size += TOT_BYTES(block_header->size);
		traverse->next = block_header->next;
		set_mem((char *)traverse,traverse->size,FREEFLAG);
		if (DEBUG_LEVEL == DEBUG_FILL)
			fill_mem((char *)ret_addr(traverse), traverse->size, 
			          FREEFLAG);		
	}
	else
		traverse->next = block_header;
	
#ifdef DEBUGIT
	cout << "lfree & traverse are " << traverse << endl;
	cout << "Size of next block on free list is "
	     << traverse->next->size << endl;
#endif		

	lfree = traverse;
}

// strdup will malloc a copy of the input string.
char *dbg_strdup(char *instr)
{
	char *outstr=0;
	
	if (!instr)
	{
		cout << "strdup: FATAL - 0 argument!\n";
		return(0);
	}
	 
	outstr = (char *) dbg_malloc(sizeof(char) *(strlen(instr) + 1));
	if (!outstr)
	{
		cout << "strdup: FATAL - malloc failed!\n";
		return(0);
	}
	
	strcpy(outstr, instr);
	
	return(outstr);
}

unsigned long calculate_free_bytes(void)
{
	headerp traverse;
	unsigned long bytes=0;
	
	if (!lfree)
	{
		return(0);
	}
	
	for (traverse = freelist.next; !(traverse == &freelist); 
	     traverse = traverse->next)
	{
		if (!check_magic(traverse,FREEFLAG))
		{
			cout << 
			"calculate_free_bytes: BAD MAGIC NUMBER! CORRUPTED HEAP!\n";
			return(0);
		} 		

		bytes += traverse->size;
	}
	
	return(bytes);
}

int calculate_largest_used(void)
{
	headerp traverse,inbetween;
	int biggest=0;
	
	for (traverse = freelist.next; !(traverse == &freelist); 
	     traverse = traverse->next)
	{

		if (!check_magic(traverse,FREEFLAG))
			return(0);
			
		inbetween = (headerp) ( (char *) traverse 
		          + TOT_BYTES(traverse->size));
		if (inbetween != traverse->next)
		{	
			while (inbetween < traverse->next)
			{
				if (inbetween->magic != MAGIC_MALLOC)
					break;
				else
				{
					if (end_magic(ret_addr(inbetween)) != MAGIC_MALLOC)
						return(0);
					if (biggest < inbetween->size)
						biggest = inbetween->size;
				}
	 			
	 			inbetween = (headerp) ( (char *) inbetween 
	 			           + TOT_BYTES(inbetween->size));
	 		} // end of while
	 	}
	 } // end of for
	 
	return(biggest);
}			

void print_stats(void)
{
	stats.free_blocks = check_free_list(&(stats.largest_free_block));
	stats.freed_bytes = calculate_free_bytes();
	stats.largest_used_block = calculate_largest_used();
	stats.total_free_mem = (unsigned long) 0;
	stats.heap_min = (long) HeapBounds.min;
	stats.heap_max = (long) HeapBounds.max;
	
	cout << "HEAP STATISTICS\n";
	cout << " - current heap size	: " << stats.current_heap_size << endl;
	cout << " - free blocks	        : " << stats.free_blocks << endl;
	cout << " - freed bytes	        : " << stats.freed_bytes << endl;
	cout << " - used blocks	        : " << stats.used_blocks << endl;
	cout << " - used bytes	        : " << stats.used_bytes << endl;
	cout << " - largest free block	: " << stats.largest_free_block << endl;
	cout << " - largest used block	: " << stats.largest_used_block << endl;
	cout << " - system memory left	: " << stats.total_free_mem << endl;
	cout << " - lowest address (dec): " << stats.heap_min << endl;
	cout << " - highest address(dec): " << stats.heap_max << endl;
	cout << " - total allocations   : " << stats.tot_allocs << endl;
	cout << " - total frees         : " << stats.tot_frees << endl;
}	

void *dbg_realloc(void *ptr, long size)
{
	headerp block_header,freeblock;
	void *newp=NULL;
	
	if (!ptr)
		return(dbg_malloc(size));
		
#ifdef DEBUGIT
	cout << "In dbg_realloc, requesting " << size << " bytes.\n";
#endif		

	block_header = (headerp) ((char *)ptr - START_BYTES);
	if (block_header->size < MULTIPLE_OF(size))
	{
#ifdef DEBUGIT
	cout << "Current size is " << block_header->size
	     << ". Mallocing larger.\n";
#endif		
		// get a larger block
		newp = dbg_malloc(size);
		
		if (!newp)
			return(NULL);
		memcpy(newp,ptr,block_header->size);
		
		dbg_free(ptr);
		return(newp);
	}
	else if (block_header->size < TOT_BYTES(size))
		return(ptr);
	else
	{
		// chop off the end of the block
		freeblock = (headerp) ((char *) block_header + TOT_BYTES(size));
		set_mem((char *)freeblock, 
		         block_header->size - TOT_BYTES(size),MALLOCFLAG);
		dbg_free(ret_addr(freeblock));
		set_mem((char *)block_header,MULTIPLE_OF(size),MALLOCFLAG);
		if (DEBUG_LEVEL == DEBUG_FILL)
			fill_mem((char *)ret_addr(block_header), block_header->size, 
			          MALLOCFLAG);		
		return(ptr);
	}
}

void dump_block(int block_number, headerp block)
{
	char dump_type,continue_char;
	unsigned char *bytep;
	int i,num_bytes;

	cout << "This block is " << block->size << " bytes.\n";
	cout << "(c)ontinue (e)xit : ";
	cin >> continue_char;

	if (continue_char == 'e')
		return;

	cout << "Dump as (c)har or (h)ex: ";
	cin >> dump_type;

	bytep = (unsigned char *) ret_addr(block);
	cout << "MEMORY DUMP OF BLOCK-NUMBER " << block_number << endl;
	cout << "**************** START ****************\n";
	num_bytes = block->size;
	switch (dump_type) {
		case 'c':
			for (i=0; i < num_bytes; i++)
			{
				if (isprint(*bytep))
					cout << *bytep;
				else
					cout << "~";

				if ((!(i % 60)) && i)
					cout << endl;
				bytep++;	
			}
			cout << endl;
			break;
		case 'h':
			for (i=0; i < num_bytes; i++)
			{
				cout << hex << (short) *bytep;
				if ((!(i % 60)) && i)
					cout << endl;
				bytep++;	
			}
			cout << endl;
			cout << dec;
			break;
		default:
			cout << "Unknown type!\n";
			return;
	};
	cout << "***************** END *****************\n";
}

void memory_dump(void)
{
	headerp traverse,inbetween;
	int cnt=0,block_num;
	int done=0;
	int blocks_seen=0;
	
	if (!lfree)
		return;

	while (!done)
	{
		cnt = 0;
		cout << "Enter the block number to dump (-1 to exit) : ";
		cin >> block_num;

		if (block_num < 1)
			return;

		for (traverse = freelist.next; !(traverse == &freelist); 
		     traverse = traverse->next)
		{
			cnt++;

			/* dump a free block */
			if (cnt == block_num)
			{
			 		dump_block(cnt,traverse);
			}
			
			inbetween = (headerp) ( (char *) traverse + 
			             TOT_BYTES(traverse->size));
			if (inbetween != traverse->next)
			{	
				while ( (inbetween < traverse->next) ||
					(blocks_seen < stats.used_blocks) )
				{
					/* if we hit a non-malloced block, which could occur if
					there is only one free block, and when the heap is not
					contiguous. get out and go to the next free block. */
					if (inbetween->magic != MAGIC_MALLOC)
						break;
					else
					{
						if (end_magic(ret_addr(inbetween)) != MAGIC_MALLOC)
						{
							cout << "memory_map: CORRUPTED HEAP!\n";
							cout << "block number: " << cnt << endl;
						}
					}
				
					cnt++;

					// dump the malloced block
					if (cnt == block_num)
					{
							dump_block(cnt,inbetween);
					}			

	 				inbetween = (headerp) ( (char *) inbetween
	 				           + TOT_BYTES(inbetween->size));
		 		} // end of while
		 	}
	 	
		 } // end of for
	} // end of while not done
}		

void *operator new(size_t num_bytes)
{
	void *p = 0;
	
	if (DEBUG_LEVEL == DEBUG_NONE)
	{
		p = malloc(num_bytes);
	}
	else if (DEBUG_LEVEL >= DEBUG_MINIMUM)
	{
		p = dbg_malloc(num_bytes);
	}
	return p;
}

#ifdef __MWERKS__
void *operator new(size_t, void *p) { return p; }
#endif

void operator delete(void *ptr)
{
	if (DEBUG_LEVEL == DEBUG_NONE)
	{
		free(ptr);
	}
	else if (DEBUG_LEVEL >= DEBUG_MINIMUM)
	{
		dbg_free(ptr);
	}
}
