// dbgmem.h
#ifndef _DEBUG_MEMORY
#define _DEBUG_MEMORY

// #define DEBUGIT  // to debug the memory debugger

// Alignment defines
struct fooalign {char x; double d;};
#define DEFAULT_ALIGNMENT ((char *) &((struct fooalign *) 0)->d - (char *) 0)
// NB: DEFAULT_ALIGNMENT macro is from GNU-C source by the Free Software Foundation

#define MULTIPLE_OF(x) ( ((x)%DEFAULT_ALIGNMENT) ? \
						((DEFAULT_ALIGNMENT - ((x)%DEFAULT_ALIGNMENT)) + (x)) : (x))

// header block
typedef struct header header;
typedef header *headerp;
struct header {
	unsigned magic;
	headerp next;
	unsigned size;
};

// pointer validation for free
struct valid {
	void *min;
	void *max;
	short firstMalloc;
} HeapBounds = { 0, 0, 1 };

// Start & End Bytes 
#define START_BYTES (MULTIPLE_OF(sizeof(header)))
#define END_BYTES (MULTIPLE_OF(sizeof(unsigned)))
// s is the number of bytes the user requested 
#define TOT_BYTES(s) (MULTIPLE_OF(s) + START_BYTES + END_BYTES)
#define END_OFFSET(s) (MULTIPLE_OF(s) + START_BYTES)
// x is a total number of bytes available, like from get_mem
#define USABLE_BYTES(x) (x - (START_BYTES + END_BYTES))

// magic numbers
#define MAGIC_MALLOC 0x1212 // only fills 2 bytes
#define MAGIC_FREE 0x2323   // only fills 2 bytes
#define MALLOC_COLOR1 0xbe
#define MALLOC_COLOR2 0xef  // 2 bytes - together are 0xbeef
#define FREE_COLOR1 0xde
#define FREE_COLOR2 0xad    // 2 bytes - together are 0xdead
#define MALLOCFLAG 1
#define FREEFLAG 0

// macros
#define ret_addr(ptr) ( (void *) ((char *)ptr + START_BYTES) )
#define header_magic(ptr) (((headerp)((char *)ptr - START_BYTES))->header_magic)
#define header_size(ptr) (((headerp)((char *)ptr - START_BYTES))->size)
#define end_magic(ptr) (* (unsigned *) ((char *)ptr + MULTIPLE_OF(header_size(ptr))))

// default sizes
#define DEFAULT_BLOCK_SIZE 4096
#define DEBUG_BOUND 128000

// free list
static header freelist; /* freelist */
static headerp lfree = NULL; /* last freed block */

// malloc statistics
struct heap_info {
	unsigned long current_heap_size;
	unsigned long free_blocks;
	unsigned long freed_bytes;
	unsigned long used_blocks;
	unsigned long used_bytes;
	unsigned largest_free_block;
	unsigned largest_used_block;
	unsigned long total_free_mem;  // platform specific
	unsigned long heap_min;
	unsigned long heap_max;
	unsigned long tot_allocs;
	unsigned long tot_frees;
} stats = {0,0,0,0,0,0,0,0,0,0,0};

// debug levels
#define DEBUG_NONE 0
#define DEBUG_MINIMUM 1  // guard blocks and label
#define DEBUG_FILL    2  // "color" in blocks

// To change debugging level add a choice here!!
#define DEBUG_LEVEL DEBUG_FILL

//#define DEBUG_LEVEL DEBUG_NONE

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL DEBUG_MINIMUM
#endif


// debug prototypes
void set_mem(char *ptr, unsigned size,short flag);
void fill_mem(char *ptr, unsigned numbytes, short color);
headerp get_mem(unsigned size);
int check_magic(headerp ptr, short flag);
int check_free_list(unsigned *largest);
void print_free_list(void);
void memory_map(void);
int check_heap(void);
void *dbg_malloc(unsigned num_bytes);
void dbg_free(void *ptr);
char *dbg_strdup(char *instr);
void *dbg_realloc(void *ptr, long size);

#endif
