/* count.cpp - this program will read in a block of text (up
to 5000 characters and count the number of characters, words,
and paragraphs in the text.  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifdef __SC__
#include <console.h>
#endif

#ifdef __TCPLUSPLUS__
#include <conio.h>
#endif

#include "count.h"

/* ***********************************************************
     FUNCTION NAME:  read_text
     PURPOSE:  gets characters from standard input until either an end of file or 5000
			     characters have been entered.
     INPUT:  storage - the character string to store the text.
     OUTPUT:  an integer which is the number of chararacters read.
     AUTHOR: MCD
    ************************************************************ */
int read_text(char *storage)
{
	char inchar;
	int i=0;
	
	while ( ((inchar = getchar()) != EOF) && (i<MAX_CHARS) )
	{
		*storage++ = inchar;
		i++;
	}

	if (i == MAX_CHARS)
		printf("\nMaximum characters (%d) reached!\n",MAX_CHARS);
	
	return(i);
}

/* ***********************************************************
     FUNCTION NAME:  count
     PURPOSE:  counts characters, words and paragraphs in the input text.
     INPUT:  storage - the character string holding the text.
			length - the length of the string.
			coords - a THINKC specific structure used to go to a specific coordinate on
					   the screen.
     OUTPUT:  counts are written to stdout.  function returns nothing.
     AUTHOR: MCD
    ************************************************************ */
/* count */
void count(char *storage, int length, count_posp coords)
{
	int i;
	score thescore= {0,0,0};
	
	short int char_flag=0;
	
	char last_char='~';
	
	/* we use cgotoxy to keep the cursor fixed at one spot so that
		the numbers will increment in the same position and not
		fill the entire screen. */
#ifdef __SC__
	cgotoxy(coords->char_pos.col,coords->char_pos.row, stdout);
#endif
#ifdef __TCPLUSPLUS__
	gotoxy(coords->char_pos.col, coords->char_pos.row);
#endif
	printf("%d",thescore.char_count);
#ifdef __SC__
	cgotoxy(coords->word_pos.col,coords->word_pos.row, stdout);
#endif
#ifdef __TCPLUSPLUS__
	gotoxy(coords->word_pos.col, coords->word_pos.row);
#endif
	printf("%d",thescore.word_count);
#ifdef __SC__
	cgotoxy(coords->par_pos.col,coords->par_pos.row, stdout);
#endif
#ifdef __TCPLUSPLUS__
	gotoxy(coords->par_pos.col, coords->par_pos.row);
#endif
	printf("%d",thescore.par_count);
	
	/* the methodology for counting properly is to go through every character.
		Save the last character looked at and by using the two figure when you
		hit the end of a word or the beginning of a paragraph. */
	for (i=0; i<length; i++)
	{
		if ( (isalnum(storage[i])) || (ispunct(storage[i])) )
		{
			if (!thescore.par_count)
			{
				thescore.par_count++;
#ifdef __SC__
				cgotoxy(coords->par_pos.col, coords->par_pos.row,stdout);
#endif
#ifdef __TCPLUSPLUS__
				gotoxy(coords->par_pos.col, coords->par_pos.row);
#endif
				printf("%d",thescore.par_count);
			}	
				
			if (i == length - 1)
			{
				/* last char is a non-whitespace,
					count last word. */
				thescore.word_count++;
#ifdef __SC__
				cgotoxy(coords->word_pos.col, coords->word_pos.row,stdout);
#endif
#ifdef __TCPLUSPLUS__
				gotoxy(coords->word_pos.col, coords->word_pos.row);
#endif
				printf("%d",thescore.word_count);
			}
			
			thescore.char_count++;
#ifdef __SC__
			cgotoxy(coords->char_pos.col,coords->char_pos.row, stdout);
#endif
#ifdef __TCPLUSPLUS__
			gotoxy(coords->char_pos.col, coords->char_pos.row);
#endif
			printf("%d",thescore.char_count);
			char_flag = 1;
		} 
		else if ( (isspace(storage[i])) && (char_flag) )
		{
			thescore.word_count++;
#ifdef __SC__
			cgotoxy(coords->word_pos.col,coords->word_pos.row,stdout);
#endif
#ifdef __TCPLUSPLUS__
			gotoxy(coords->word_pos.col, coords->word_pos.row);
#endif
			printf("%d",thescore.word_count);
			char_flag = 0;
		}
	
		if ( ((storage[i] == '\n') && (last_char == '\n')) ||
		      (storage[i] == '\t')  )
		{
			thescore.par_count++;
#ifdef __SC__
			cgotoxy(coords->par_pos.col,coords->par_pos.row,stdout);
#endif
#ifdef __TCPLUSPLUS__
			gotoxy(coords->par_pos.col, coords->par_pos.row);
#endif
			printf("%d",thescore.par_count);
		}
		last_char = storage[i];
	} /* for */
}

/* ***********************************************************
     FUNCTION NAME:  main for count.c
     PURPOSE:  calls read_text, prints a header, then calls count to print the counts.
     INPUT:  none.
     OUTPUT:  none.
     AUTHOR: MCD
    ************************************************************ */
void main()
{
	char text[MAX_CHARS];
	int txt_length = 0;
	int x,y;
	
	count_pos positions;
	
	// set up console size.  This is specific to THINK C. NON-ANSI
#ifdef __SC__
	cshow(stdout);
#endif
	
#ifdef __SC__
	// cgotoxy is similar to TURBO C's gotoxy
	cgotoxy(1,1,stdout);
#endif

#ifdef __TCPLUSPLUS__
	gotoxy(1,1);
#endif
	
	printf("Type text and hit ctrl-D or command-D to stop.\n");
	txt_length = read_text(text);
	
	printf("+------------- COUNT ---------------+\n");
	/* now count the characters */
	/* cgetxy is similar to _gettextposition, it gets the
	current position of the cursor. NON-ANSI.*/
#ifdef __SC__
	cgetxy(&x,&y,stdout);
#endif

#ifdef __TCPLUSPLUS__
	x = wherex();
	y = wherey();
#endif
	
	positions.char_pos.row = y;
	positions.word_pos.row = y+1;
	positions.par_pos.row = y+2;
	positions.char_pos.col = 19;
	positions.word_pos.col = 19;
	positions.par_pos.col = 19;
	
	printf("# of characters : \n");
	printf("# of words      : \n");
	printf("# of paragraphs : \n");
	
	count(text,txt_length,&positions);
}
