// PPbuf.cpp

#include <iostream.h>
#include <fstream.h>
#include <assert.h>
#include <string.h>
#include <new.h>

#include "PPbuf.h"

error PPbuf::PPerror;  // static error object

// no argument constructor
PPbuf::PPbuf() 
{
	_len = 0;
	terminator = '\n';
	_Space = CHUNKSZ;
	_Strings = new String[CHUNKSZ];
	if (!_Strings)
		PPerror.StandardError(NOMEM,__FILE__,__LINE__);
}

// when we need to expand the size of the buffer
void PPbuf::GrowBuf()
{
	String *tmp=0;
	int i=0;

	tmp = new String[_Space + CHUNKSZ];
	if (!tmp)
		PPerror.StandardError(NOMEM,__FILE__,__LINE__);
		
	// you cannot do the memcpy because
	// I must call delete []
	for (i=0; i < _Space; i++)
		tmp[i] = _Strings[i];
	
	// arrays are allocated differently. if you allocate an
	// array with new[] you MUST free it with delete [].
	delete [] _Strings;
	
	_Strings = tmp;
	_Space += CHUNKSZ;
}

// destructor
PPbuf::~PPbuf()
{
	int i=0;
		
	delete []_Strings;
	
	// destructor called automatically for each object in
	// the array!
}

// one argument constructor
// initialize PPbuf object from an "unprotected" ppbuf
// unprotected ppbuf should be NULL-terminated
PPbuf::PPbuf(char **ppbuf, int pplen)
{
	int i=0;

	assert(ppbuf!=NULL);
		
	// create a new PPbuf based on the length of ppbuf
	if (!pplen)
		for (i=0; ppbuf[i]; i++, pplen++) ;
	
	_Strings = new String[pplen + 1];
	if (!_Strings)
		PPerror.StandardError(NOMEM,__FILE__,__LINE__);
		
	if (ppbuf)
	{
		for (i=0; i < pplen; i++)
		{
			_Strings[i] = ppbuf[i];
		}
	}
	_Strings[pplen] = 0;  // NULL terminate
	_len = pplen;
	_Space = pplen + 1;
	terminator = '\n';
}

// initialize PPbuf from a file.  replaced "file2ppbuf"
PPbuf::PPbuf(char *filename)
{
	char buf[256];
	ifstream infile(filename);
	int i,count;
	
	if (!infile)
		PPerror.StandardError(NOTOPEN,__FILE__,__LINE__);
	
	// count the lines in the file
	for (count = 0; infile; count++)
		infile.getline(buf,256);
		
	// now get space for PPbuf
	_len = count;
	count++;
	_Strings = new String[count];
	if (!_Strings)
		PPerror.StandardError(NOMEM,__FILE__,__LINE__);		
		
	// close and reopen file
	infile.close();
	infile.open(filename);
	if (!infile)
		PPerror.StandardError(NOREWIND,__FILE__,__LINE__);
	
	// Fill PPbuf
	i=0;
	while (infile)
	{
		infile.getline(buf,256);
		_Strings[i] = buf;
		i++;
	}
	
	_Strings[_len] = 0;   // NULL terminate
	terminator = '\n';
}

// this constructor replaces toks_from_str()
PPbuf::PPbuf(String line, char *delimiters)
{
  	char *temp=0, *tok=0;
    int cnt=0;
   	
   	if (!line || !delimiters)
		PPerror.StandardError(NULLARG,__FILE__,__LINE__);

	// initialize PPbuf
	_len = 0;
	terminator = '\n';
	_Space = CHUNKSZ;
	_Strings = new String[CHUNKSZ];
	if (!_Strings)
		PPerror.StandardError(NOMEM,__FILE__,__LINE__);
  
    temp = line.Strdup();
    tok = strtok(temp,delimiters);
    if (tok) _Strings[cnt] = tok; _len++;
    cnt++;
   
    while ( (tok = strtok(NULL,delimiters)) != NULL)
    {
    	if (_len == _Space)
    		this->GrowBuf();

   	    _Strings[cnt] = tok; _len++;
	    cnt++;
    } /* while strtok does not return NULL */

	delete temp; 
}

ostream &operator << (ostream &stream, PPbuf &obj)
{
	int i=0;
	
	// dump the ppbuf to the stream with the terminator.
	for (i=0; i < obj._len; i++)
	{
		stream << obj._Strings[i];
		if (obj.terminator == '\n') stream << endl;
	}
	
	return stream;
}

// copy constructor
PPbuf::PPbuf(PPbuf &other)
{
	int i=0;

	_len = other._len;
	_Space = other._Space;
	terminator = other.terminator;
	
	// create a new PPbuf	
	_Strings = new String[_len + 1];
	if (!_Strings)
		PPerror.StandardError(NOMEM,__FILE__,__LINE__);
		
	if (other._Strings)
	{
		for (i=0; i < _len; i++)
		{
			_Strings[i] = other._Strings[i];
		}
	}
	_Strings[_len] = 0;  // NULL terminate
}

// overload the equal operator
PPbuf &PPbuf::operator = (PPbuf &other)
{
	int i=0;
	
	if (this == &other)  // self-assignment
		return *this;
	
	_len = other._len;
	_Space = other._Space;
	terminator = other.terminator;

	// if previous delete it
	if (_Strings) delete [] _Strings;
		
	// create a new PPbuf	
	_Strings = new String[_len + 1];
	if (!_Strings)
		PPerror.StandardError(NOMEM,__FILE__,__LINE__);
		
	if (other._Strings)
	{
		for (i=0; i < _len; i++)
		{
			_Strings[i] = other._Strings[i];
		}
	}
	_Strings[_len] = 0;  // NULL terminate
	return *this;
}

// overload the conversion operator when using with C routines
PPbuf::operator char **()
{
	int i;
	char **outppbuf = NULL;
	
	outppbuf = new char *[_len + 1];
	if (!outppbuf)
		PPerror.StandardError(NOMEM,__FILE__,__LINE__);

	for (i=0; i < _len; i++)
		outppbuf[i] = _Strings[i].Strdup();
	
	outppbuf[_len] = 0;
	return outppbuf;
}

// overload the bracket operator
String &PPbuf::operator[](int index)
{
	assert(index < _len);
	assert(index >= 0);
	return(_Strings[index]);
}

istream &operator >> (istream &stream, PPbuf &obj)
{
	int i=0;
	char buf[256];
	
	while (stream)
	{
		if (i == obj._Space)
			obj.GrowBuf();
		stream.getline(buf,256);
		obj._Strings[i] = buf;
		i++;
	}
	
	obj._len = i;
	obj._Strings[obj._len] = 0;   // NULL terminate
	obj.terminator = '\n';
	return stream;
}

int PPbuf::size()
{
	return _len;
}

void PPbuf::AppendString(String instr)
{
	if (_len == _Space)
		this->GrowBuf();
	
	_Strings[_len] = instr;
	_len++;
}

void PPbuf::AppendString(char *instr)
{
	if (_len == _Space)
		this->GrowBuf();
	
	_Strings[_len] = instr;
	_len++;	
}

// start is the index of the ppbuf to start with
// len is the number of lines to copy
PPbuf PPbuf::dupPPbuf(int start, int len)
{
	int i=0,j=0;
	PPbuf temp;
	
	assert(len < _len);
	assert(start >= 0);
	
	for (i=start,j=0; i < _len && j < len; i++,j++)
		temp.AppendString(_Strings[i]);
	
	return temp;
}

String PPbuf::RemoveFirst()
{
	String tmp;
	String *pptmp;

	if (!_len)
		return 0;
			
	pptmp = new String[_len];
	if (!pptmp)
		PPerror.StandardError(NOMEM,__FILE__,__LINE__);
	
	tmp = _Strings[0];
	memcpy((void *)pptmp, (void *)(_Strings+1), (sizeof(String) * _len));
	delete _Strings; // do not call the destructor for each object
	_Strings = pptmp;
	_len--;
	return tmp;
}

String PPbuf::RemoveLast()
{
	String tmp;
	tmp = _Strings[_len-1];
	_len--;
	return tmp;
}

